# Hide pygame support prompt
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'


from gymnasium.envs.registration import register


def register_highway_envs():
    for spots_in_row in [10, 12, 14, 16]:
        for goal_index, parked_cars in [
            (1, ()),
            (2, ()),
            (3, ()),
            (4, ()),
            (5, ()),
            (6, ()),
            (7, ()),
            (8, ()),
            (9, ()),
            (10, ()),
            (11, ()),
            (12, ()),
            (13, ()),
            (14, ()),
            (15, ()),
            (16, ()),
            (17, ()),
            (18, ()),
            (19, ()),
            (20, ()),
            (21, ()),
            (22, ()),
            (23, ()),
            (24, ()),
            (2, (0, 1, 3, 4, 5, 7)),
            (6, (0, 1, 3, 4, 5, 7)),
            (8, (0, 1, 3, 4, 5, 7))
        ]:
            env_id = f'Parking-S-{spots_in_row}-PC-{"Y".join([str(pc) for pc in parked_cars])}-GI-{goal_index}-v0'
            register(
                id=env_id,
                entry_point='gr_envs.highway_scripts.envs.parking_env:ParkingEnv',
                kwargs={
                    "n_spots": spots_in_row,
                    "goal_index": goal_index,
                    "parked_cars": parked_cars
                },
            )
            gc_env_id = f'Parking-S-{spots_in_row}-PC-{"Y".join([str(pc) for pc in parked_cars])}-v0'
            register(
                id=gc_env_id,
                entry_point='gr_envs.highway_scripts.envs.parking_env:ParkingEnv',
                kwargs={
                    "n_spots": spots_in_row,
                    "parked_cars": parked_cars
                },
            )

register_highway_envs()



import gymnasium
from gymnasium.envs.registration import register
from gr_envs.maze_scripts.envs.maze.generate_maze import gen_empty_env, gen_four_rooms_env, gen_maze_with_obstacles
from gymnasium_robotics.core import GoalEnv
from gymnasium_robotics.envs.maze import maps


def register_robotics_envs():
    """Register all environment ID's to Gymnasium."""
    ### MAZE SPECIAL ENVS ###
    for reward_type in ["sparse", "dense"]:
        suffix = "Dense" if reward_type == "dense" else ""
        for width, height in [(11, 11)]:
            for start_x, start_y in [(1, 1)]:
                register(
                    id=f"PointMaze-FourRoomsEnv{suffix}-{width}x{height}-Goals-9x1-1x9-9x9",
                    entry_point="gymnasium_robotics.envs.maze.point_maze:PointMazeEnv",
                    kwargs= {
                        "reward_type": reward_type,
                        "maze_map": gen_four_rooms_env(width, height, [(start_x, start_y)], [(1, 9), (9, 1), (9, 9)]),
                        "continuing_task": False
                    },
                    max_episode_steps=900,
                )
            for goal_x, goal_y in [(1, 9), (9, 1), (9, 9), (7, 3), (3, 7), (6, 4), (4, 6), (3, 3), (6, 6), (4, 4),
                                   (3, 4), (7, 7), (6, 7), (8, 8), (7, 4), (4, 7), (6, 3), (3, 6), (5, 5), (5, 1),
                                   (1, 5), (8, 2), (2, 8), (3, 4), (4, 3)]:
                register(
                    id=f"PointMaze-EmptyEnv{suffix}-{width}x{height}-Goal-{goal_x}x{goal_y}",
                    entry_point="gymnasium_robotics.envs.maze.point_maze:PointMazeEnv",
                    kwargs= {
                        "reward_type": reward_type,
                        "maze_map": gen_empty_env(width, height, [(start_x, start_y)], [(goal_x, goal_y)]),
                        "continuing_task": False
                    },
                    max_episode_steps=900,
                )
                register(
                    id=f"PointMaze-FourRoomsEnv{suffix}-{width}x{height}-Goal-{goal_x}x{goal_y}",
                    entry_point="gymnasium_robotics.envs.maze.point_maze:PointMazeEnv",
                    kwargs= {
                        "reward_type": reward_type,
                        "maze_map": gen_four_rooms_env(width, height, [(start_x, start_y)], [(goal_x, goal_y)]),
                        "continuing_task": False
                    },
                    max_episode_steps=900,
                )
                register(
                    id=f"PointMaze-ObstaclesEnv{suffix}-{width}x{height}-Goal-{goal_x}x{goal_y}",
                    entry_point="gymnasium_robotics.envs.maze.point_maze:PointMazeEnv",
                    kwargs= {
                        "reward_type": reward_type,
                        "maze_map": gen_maze_with_obstacles(11, 11, [(1, 1)], [(goal_x, goal_y)], [(2, 2), (2, 3), (2, 4), (3, 2), (3, 3), (3, 4), (4, 2), (4, 3), (4, 4)]),
                        "continuing_task": False
                    },
                    max_episode_steps=900,
                )
    ### END OF MAZE SPECIAL ENVS ###
    
    ### kitchen special envs ###
    register(
        id="FrankaKitchen-v1",
        entry_point="gymnasium_robotics.envs.franka_kitchen:KitchenEnv",
        max_episode_steps=1000, # default is 280, but since this is a sparse reward environment, better increase it so the agent can 
    )
    ### END OF kitchen special envs ###

    def _merge(a, b):
        a.update(b)
        return a

register_robotics_envs()


from gymnasium.envs.registration import register
from minigrid import minigrid_env, wrappers
from minigrid.core import roomgrid
from minigrid.core.world_object import Wall, Lava


def register_minigrid_envs():
    # Dyanmic Goal Empty Custom 13
    # base goals:
    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-11x1-v0", # for base
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (11, 1), "obstacle_type": Wall, "start_pos": (1, 1)},
    )
    
    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-11x11-v0", # for dynamic
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (11, 11), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-1x11-v0", # for base
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (1, 11), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-7x11-v0", # for base
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (7, 11), "obstacle_type": Wall, "start_pos": (1, 1)},
    )
    # end of base goals
    
    #dynamic goals:
    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-11x3-v0", # for dynamic
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (11, 3), "obstacle_type": Wall, "start_pos": (1, 1)},
    )
    
    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-11x5-v0", # for dynamic
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (11, 5), "obstacle_type": Wall, "start_pos": (1, 1)},
    )
    
    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-11x8-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (11, 8), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-8x1-v0", # for base
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (8, 1), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-6x1-v0", # for dynamic
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (6, 1), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-1x7-v0", # for dynamic, to confuse with 9x1(within same room)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (1, 7), "obstacle_type": Wall, "start_pos": (1, 1)},
    )
    
    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-5x9-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (5, 9), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-10x6-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (10, 6), "obstacle_type": Wall, "start_pos": (1, 1)},
    )

    register(
        id="MiniGrid-SimpleCrossingS13N4-DynamicGoal-6x9-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 13, "num_crossings": 4, "goal_pos": (6, 9), "obstacle_type": Wall, "start_pos": (1, 1)},
    )
    # end of dynamic goals

    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-7x7-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (7, 7), "obstacle_type": Lava},
    )

    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-1x7-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (1, 7), "obstacle_type": Lava},
    )

    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-7x1-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (7, 1), "obstacle_type": Lava},
    )

    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-1x3-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (1, 3), "obstacle_type": Lava},
    )
    
    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-6x5-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (6, 5), "obstacle_type": Lava},
    )
    
    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-4x7-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (4, 7), "obstacle_type": Lava},
    )
    
    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-2x5-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (2, 5), "obstacle_type": Lava},
    )
    
    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-5x2-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (5, 2), "obstacle_type": Lava},
    )
    
    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-4x5-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (4, 5), "obstacle_type": Lava},
    )
    
    register(
        id="MiniGrid-LavaCrossingS9N2-DynamicGoal-1x1-v0", # for dynamic, to confuse with 9x1 (outside the room of 9x1)
        entry_point="gr_envs.minigrid_scripts.envs:DynamicGoalCrossingCustom13Env",
        kwargs={"size": 9, "num_crossings": 2, "goal_pos": (1, 1), "obstacle_type": Lava},
    )
    

register_minigrid_envs()
# python experiments.py --recognizer graml --domain point_maze --task L2 --partial_obs_type continuing --point_maze_env obstacles --collect_stats --inference_same_seq_len



# Additions not in minigrid:
from gr_envs.minigrid_scripts.envs.dynamic_goal_crossing import DynamicGoalCrossingEnv
from gr_envs.minigrid_scripts.envs.dynamic_goal_crossing_custom import DynamicGoalCrossingCustom13Env
from gr_envs.minigrid_scripts.envs.custom_color import CustomColorEnv
from gr_envs.minigrid_scripts.envs.dynamic_goal_empty import DynamicGoalEmpty
from gr_envs.minigrid_scripts.envs.dynamic_four_rooms import DynamicFourRoomsEnv



import gymnasium as gym
from gymnasium_robotics.envs.maze.maps import R, G

def gen_empty_env(width, height, initial_states, goal_states):
    # Create an empty environment matrix with walls (1) around the edges
    env = [[1 if x == 0 or x == width-1 or y == 0 or y == height-1 else 0 for x in range(width)] for y in range(height)]

    # Place initial states (R) and goal states (G)
    for x, y in initial_states:
        env[y][x] = R
    for x, y in goal_states:
        env[y][x] = G
    
    return env

def gen_four_rooms_env(width, height, initial_states, goal_states):
    # Create an empty environment matrix with walls (1) around the edges
    env = [[1 if x == 0 or x == width-1 or y == 0 or y == height-1 else 0 for x in range(width)] for y in range(height)]
    
    # Add walls for the four rooms structure
    for y in range(1, height-1):
        env[y][width // 2] = 1 if y != height // 4 and y != height * 3 // 4 else 0
    for x in range(1, width-1):
        env[height // 2][x] = 1 if x != width // 4 and x != width * 3 // 4 else 0
    
    # Place initial states (R) and goal states (G)
    for x, y in initial_states:
        env[y][x] = R
    for x, y in goal_states:
        env[y][x] = G
    
    return env

def gen_maze_with_obstacles(width, height, initial_states, goal_states, obstacles):
    # Create an empty environment matrix with walls (1) around the edges
    env = [[1 if x == 0 or x == width-1 or y == 0 or y == height-1 else 0 for x in range(width)] for y in range(height)]
    
    # Place obstacles (1)
    for x, y in obstacles:
        env[y][x] = 1
    
    # Place initial states (R) and goal states (G)
    for x, y in initial_states:
        env[y][x] = R
    for x, y in goal_states:
        env[y][x] = G
    
    return env

# Example usage
if __name__ == "__main__":
    width, height = 9, 9
    initial_states = [(1, 1)]
    goal_states = [(1, 7)]
    obstacles = [(3, 1), (3, 2), (3, 3)]
    
    empty_env = gen_empty_env(width, height, initial_states, goal_states)
    for row in empty_env:
        print(row)
    
    print()
    
    four_rooms_env = gen_four_rooms_env(width, height, initial_states, goal_states)
    for row in four_rooms_env:
        print(row)
    
    print()
    
    maze_with_obstacles = gen_maze_with_obstacles(width, height, initial_states, goal_states, obstacles)
    for row in maze_with_obstacles:
        print(row)



import os

from gymnasium.envs.registration import register

# with open(os.path.join(os.path.dirname(__file__), "version.txt"), "r") as file_handler:
#     __version__ = file_handler.read().strip()

for reward_type in ["sparse", "dense"]:
    for control_type in ["ee", "joints"]:
        reward_suffix = "Dense" if reward_type == "dense" else ""
        control_suffix = "Joints" if control_type == "joints" else ""
        kwargs = {"reward_type": reward_type, "control_type": control_type}

#         register(
#             id="PandaReach{}{}-v3".format(control_suffix, reward_suffix),
#             entry_point="panda_gym.envs:PandaReachEnv",
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )

#         register(
#             id="PandaPush{}{}-v3".format(control_suffix, reward_suffix),
#             entry_point="panda_gym.envs:PandaPushEnv",
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )

#         register(
#             id="PandaSlide{}{}-v3".format(control_suffix, reward_suffix),
#             entry_point="panda_gym.envs:PandaSlideEnv",
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )

#         register(
#             id="PandaPickAndPlace{}{}-v3".format(control_suffix, reward_suffix),
#             entry_point="panda_gym.envs:PandaPickAndPlaceEnv",
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )

#         register(
#             id="PandaStack{}{}-v3".format(control_suffix, reward_suffix),
#             entry_point="panda_gym.envs:PandaStackEnv",
#             kwargs=kwargs,
#             max_episode_steps=100,
#         )

#         register(
#             id="PandaFlip{}{}-v3".format(control_suffix, reward_suffix),
#             entry_point="panda_gym.envs:PandaFlipEnv",
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )

        register(
            id="PandaMyReach{}{}-v3".format(control_suffix, reward_suffix),
            entry_point="gr_envs.panda_scripts.envs:PandaMyReachEnv",
            kwargs={**kwargs, "goal_range_negative": -0.5, "goal_range_positive": 0.2},
            max_episode_steps=101,
        )

reward_type = "dense"
control_type = "ee"
goals = [(-0.5, -0.5, 0.1), (-0.3, -0.3, 0.1), (-0.1, -0.1, 0.1), (-0.5, 0.2, 0.1), (-0.3, 0.2, 0.1), (-0.1, 0.1, 0.1), (0.2, -0.2, 0.1), (0.2, -0.3, 0.1), (0.1, -0.1, 0.1), (0.2, 0.2, 0.1), (0.0, 0.0, 0.1), (0.1, 0.1, 0.1)]
reward_suffix = "Dense" if reward_type == "dense" else ""
control_suffix = "Joints" if control_type == "joints" else ""
for goal in goals:
    goal_str = 'X'.join([str(float(g)).replace(".", "y").replace("-","M") for g in goal])
    goal_x, goal_y, goal_z = goal
    kwargs = {"reward_type": reward_type, "control_type": control_type, "goal_x": goal_x, "goal_y": goal_y, "goal_z": goal_z}
    env_id = "PandaMyReach{}{}X{}-v3".format(control_suffix, reward_suffix, goal_str)
    #print(f"{goal}:{env_id}")
    register(
        id=env_id,
        entry_point="gr_envs.panda_scripts.envs:PandaMyReachEnv",
        kwargs=kwargs,
        max_episode_steps=101, 
    )



import itertools as itt

import numpy as np

from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Goal, Lava
from minigrid.minigrid_env import MiniGridEnv

class DynamicGoalCrossingCustom13Env(MiniGridEnv):

    """
    ### Description

    Depending on the `obstacle_type` parameter:
    - `Lava` - The agent has to reach the green goal square on the other corner
        of the room while avoiding rivers of deadly lava which terminate the
        episode in failure. Each lava stream runs across the room either
        horizontally or vertically, and has a single crossing point which can be
        safely used; Luckily, a path to the goal is guaranteed to exist. This
        environment is useful for studying safety and safe exploration.
    - otherwise - Similar to the `LavaCrossing` environment, the agent has to
        reach the green goal square on the other corner of the room, however
        lava is replaced by walls. This MDP is therefore much easier and maybe
        useful for quickly testing your algorithms.

    ### Mission Space
    Depending on the `obstacle_type` parameter:
    - `Lava` - "avoid the lava and get to the green goal square"
    - otherwise - "find the opening and get to the green goal square"

    ### Action Space

    | Num | Name         | Action       |
    |-----|--------------|--------------|
    | 0   | left         | Turn left    |
    | 1   | right        | Turn right   |
    | 2   | forward      | Move forward |
    | 3   | pickup       | Unused       |
    | 4   | drop         | Unused       |
    | 5   | toggle       | Unused       |
    | 6   | done         | Unused       |

    ### Rewards

    A reward of '1' is given for success, and '0' for failure.

    ### Termination

    The episode ends if any one of the following conditions is met:

    1. The agent reaches the goal.
    2. Timeout (see `max_steps`).

    ### Registered Configurations

    S: size of the map SxS.
    N: number of valid crossings across lava or walls from the starting position
    to the goal

    - Wall :
        - `MiniGrid-SimpleCrossingS9N1-v0`
        - `MiniGrid-SimpleCrossingS9N2-v0`
        - `MiniGrid-SimpleCrossingS9N3-v0`
        - `MiniGrid-SimpleCrossingS11N5-v0`
        - `MiniGrid-SimpleCrossingS13N3-v0`

    """

    def __init__(self, size=13, num_crossings=4, obstacle_type=Lava, start_pos = (3, 1), goal_pos =(1, 1), **kwargs):
        self.size = size
        self.num_crossings = num_crossings
        self.obstacle_type = obstacle_type

        if obstacle_type == Lava:
            mission_space = MissionSpace(
                mission_func=lambda: "avoid the lava and get to the green goal square"
            )
        else:
            mission_space = MissionSpace(
                mission_func=lambda: "find the opening and get to the green goal square"
            )

        super().__init__(
            mission_space=mission_space,
            grid_size=size,
            max_steps=4 * size * size,
            # Set this to True for maximum speed
            see_through_walls=False,
            **kwargs
        )
        self.goal_x, self.goal_y = goal_pos
        self.start_pos = start_pos

    def _gen_grid(self, width, height):
        assert width % 2 == 1 and height % 2 == 1  # odd size

        # Create an empty grid
        self.grid = Grid(width, height)

        # Generate the surrounding walls
        self.grid.wall_rect(0, 0, width, height)

        # Place the agent in the top-left corner
        self.agent_pos = np.array(self.start_pos)
        self.agent_dir = 0

        # Place a goal square in the specified location
        self.put_obj(Goal(), self.goal_x, self.goal_y)

        if self.size == 13:
            # room
            self.put_obj(self.obstacle_type(), 7, 1)
            self.put_obj(self.obstacle_type(), 7, 2)
            self.put_obj(self.obstacle_type(), 7, 3)
            self.put_obj(self.obstacle_type(), 7, 4)
            
            self.put_obj(self.obstacle_type(), 11, 4)
            self.put_obj(self.obstacle_type(), 10, 4)
            self.put_obj(self.obstacle_type(), 9, 4)
            # end_of_room
            
            self.put_obj(self.obstacle_type(), 8, 11)
            self.put_obj(self.obstacle_type(), 8, 10)
            self.put_obj(self.obstacle_type(), 8, 9)
            self.put_obj(self.obstacle_type(), 8, 8)

            self.put_obj(self.obstacle_type(), 2, 8)
            self.put_obj(self.obstacle_type(), 3, 8)
            self.put_obj(self.obstacle_type(), 4, 8)
            self.put_obj(self.obstacle_type(), 5, 8)
        
        else:
            self.put_obj(self.obstacle_type(), 2, 1)
            self.put_obj(self.obstacle_type(), 2, 2)
            self.put_obj(self.obstacle_type(), 2, 3)
            self.put_obj(self.obstacle_type(), 2, 4)
            
            self.put_obj(self.obstacle_type(), 5, 7)
            self.put_obj(self.obstacle_type(), 5, 6)
            self.put_obj(self.obstacle_type(), 5, 4)
            self.put_obj(self.obstacle_type(), 5, 3)

        self.mission = (
            "get to the green goal square"
            if self.obstacle_type == Lava
            else "find the opening and get to the green goal square"
        )



from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Goal
from minigrid.minigrid_env import MiniGridEnv


class DynamicGoalEmpty(MiniGridEnv):
    """
    ### Description

    This environment is an empty room, and the goal of the agent is to reach the
    green goal square, which provides a sparse reward. A small penalty is
    subtracted for the number of steps to reach the goal. This environment is
    useful, with small rooms, to validate that your RL algorithm works
    correctly, and with large rooms to experiment with sparse rewards and
    exploration. The random variants of the environment have the agent starting
    at a random position for each episode, while the regular variants have the
    agent always starting in the corner opposite to the goal.

    ### Mission Space

    "get to the green goal square"

    ### Action Space

    | Num | Name         | Action       |
    |-----|--------------|--------------|
    | 0   | left         | Turn left    |
    | 1   | right        | Turn right   |
    | 2   | forward      | Move forward |
    | 3   | pickup       | Unused       |
    | 4   | drop         | Unused       |
    | 5   | toggle       | Unused       |
    | 6   | done         | Unused       |

    ### Observation Encoding

    - Each tile is encoded as a 3 dimensional tuple:
        `(OBJECT_IDX, COLOR_IDX, STATE)`
    - `OBJECT_TO_IDX` and `COLOR_TO_IDX` mapping can be found in
        [minigrid/minigrid.py](minigrid/minigrid.py)
    - `STATE` refers to the door state with 0=open, 1=closed and 2=locked

    ### Rewards

    A reward of '1' is given for success, and '0' for failure.

    ### Termination

    The episode ends if any one of the following conditions is met:

    1. The agent reaches the goal.
    2. Timeout (see `max_steps`).

    ### Registered Configurations

    - `MiniGrid-Empty-5x5-v0`
    - `MiniGrid-Empty-Random-5x5-v0`
    - `MiniGrid-Empty-6x6-v0`
    - `MiniGrid-Empty-Random-6x6-v0`
    - `MiniGrid-Empty-8x8-v0`
    - `MiniGrid-Empty-16x16-v0`

    """

    def __init__(self, size=8, agent_start_pos=(1, 1), agent_start_dir=0, goal_pos =(1, 1), **kwargs):
        self.agent_start_pos = agent_start_pos
        self.agent_start_dir = agent_start_dir

        mission_space = MissionSpace(
            mission_func=lambda: "get to the green goal square"
        )

        super().__init__(
            mission_space=mission_space,
            grid_size=size,
            max_steps=4 * size * size,
            # Set this to True for maximum speed
            see_through_walls=True,
            **kwargs
        )
        self.goal_x, self.goal_y = goal_pos
        #self.reward_range = (-1, 1)
        

    def _gen_grid(self, width, height):
        # Create an empty grid
        self.grid = Grid(width, height)

        # Generate the surrounding walls
        self.grid.wall_rect(0, 0, width, height)

        # Place a goal square in the bottom-right corner
        self.put_obj(Goal(), self.goal_x, self.goal_y)

        # Place the agent
        if self.agent_start_pos is not None:
            self.agent_pos = self.agent_start_pos
            self.agent_dir = self.agent_start_dir
        else:
            self.place_agent()

        self.mission = "get to the green goal square"



from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Goal
from minigrid.minigrid_env import MiniGridEnv


class DynamicFourRoomsEnv(MiniGridEnv):

    """
    ### Description

    Classic four room reinforcement learning environment. The agent must
    navigate in a maze composed of four rooms interconnected by 4 gaps in the
    walls. To obtain a reward, the agent must reach the green goal square. Both
    the agent and the goal square are randomly placed in any of the four rooms.

    ### Mission Space

    "reach the goal"

    ### Action Space

    | Num | Name         | Action       |
    |-----|--------------|--------------|
    | 0   | left         | Turn left    |
    | 1   | right        | Turn right   |
    | 2   | forward      | Move forward |
    | 3   | pickup       | Unused       |
    | 4   | drop         | Unused       |
    | 5   | toggle       | Unused       |
    | 6   | done         | Unused       |

    ### Observation Encoding

    - Each tile is encoded as a 3 dimensional tuple:
        `(OBJECT_IDX, COLOR_IDX, STATE)`
    - `OBJECT_TO_IDX` and `COLOR_TO_IDX` mapping can be found in
        [minigrid/minigrid.py](minigrid/minigrid.py)
    - `STATE` refers to the door state with 0=open, 1=closed and 2=locked

    ### Rewards

    A reward of '1' is given for success, and '0' for failure.

    ### Termination

    The episode ends if any one of the following conditions is met:

    1. The agent reaches the goal.
    2. Timeout (see `max_steps`).

    ### Registered Configurations

    - `MiniGrid-FourRooms-v0`

    """

    def __init__(self, agent_pos=None, goal_pos=None, **kwargs):
        self._agent_default_pos = agent_pos
        self._goal_default_pos = goal_pos

        self.size = 19
        mission_space = MissionSpace(mission_func=lambda: "reach the goal")

        super().__init__(
            mission_space=mission_space,
            width=self.size,
            height=self.size,
            max_steps=100,
            **kwargs
        )

    def _gen_grid(self, width, height):
        # Create the grid
        self.grid = Grid(width, height)

        # Generate the surrounding walls
        self.grid.horz_wall(0, 0)
        self.grid.horz_wall(0, height - 1)
        self.grid.vert_wall(0, 0)
        self.grid.vert_wall(width - 1, 0)

        room_w = width // 2
        room_h = height // 2

        # For each row of rooms
        for j in range(0, 2):

            # For each column
            for i in range(0, 2):
                xL = i * room_w
                yT = j * room_h
                xR = xL + room_w
                yB = yT + room_h

                # Bottom wall and door
                if i + 1 < 2:
                    self.grid.vert_wall(xR, yT, room_h)
                    pos = (xR, (yT + 1 + yB) // 2)
                    self.grid.set(*pos, None)

                # Bottom wall and door
                if j + 1 < 2:
                    self.grid.horz_wall(xL, yB, room_w)
                    pos = ((xL + 1 + xR) // 2, yB)
                    self.grid.set(*pos, None)

        # Randomize the player start position and orientation
        if self._agent_default_pos is not None:
            self.agent_pos = self._agent_default_pos
            self.grid.set(*self._agent_default_pos, None)
            # assuming random start direction
            self.agent_dir = 2
        else:
            self.place_agent()

        if self._goal_default_pos is not None:
            goal = Goal()
            self.put_obj(goal, *self._goal_default_pos)
            goal.init_pos, goal.cur_pos = self._goal_default_pos
        else:
            self.place_obj(Goal())



from typing import Any, Dict, Union

import numpy as np

from panda_gym.envs.core import Task
from panda_gym.utils import distance


class MyReach(Task):
    def __init__(
        self,
        sim,
        get_ee_position,
        reward_type="sparse",
        distance_threshold=0.04,
        goal_x: float = None,
        goal_y: float = None,
        goal_z: float = None,
        goal_range_negative: float = -0.5,
        goal_range_positive: float = 0.2
    ) -> None:
        super().__init__(sim)
        self.reward_type = reward_type
        self.distance_threshold = distance_threshold
        self.get_ee_position = get_ee_position
        self.goal_range_low = np.array([goal_range_negative, goal_range_negative, 0.10])
        self.goal_range_high = np.array([goal_range_positive, goal_range_positive, 0.15])
        
        self._goal_x = goal_x
        self._goal_y = goal_y
        self._goal_z = goal_z
         
        with self.sim.no_rendering():
            self._create_scene()
            self.sim.place_visualizer(target_position=np.zeros(3), distance=0.9, yaw=45, pitch=-30)
            
    def _create_scene(self) -> None:
        self.sim.create_plane(z_offset=-0.4)
        self.sim.create_table(length=1.1, width=0.7, height=0.4, x_offset=-0.3)
        self.sim.create_sphere(
            body_name="target",
            radius=0.02,
            mass=0.0,
            ghost=True,
            position=np.zeros(3),
            rgba_color=np.array([0.1, 0.9, 0.1, 0.3]),
        )

    def get_obs(self) -> np.ndarray:
        return np.array([])  # no tasak-specific observation

    def get_achieved_goal(self) -> np.ndarray:
        ee_position = np.array(self.get_ee_position())
        return ee_position

    def reset(self) -> None:
        self.goal = self._sample_goal()
        self.sim.set_base_pose("target", self.goal, np.array([0.0, 0.0, 0.0, 1.0]))

    def _sample_goal(self) -> np.ndarray:
        """Randomize goal."""
        if self._goal_x == None or self._goal_y == None or self._goal_z == None:
          goal = self.np_random.uniform(self.goal_range_low, self.goal_range_high)
        else:
          goal = np.array(
            [self._goal_x, self._goal_y, self._goal_z]
          ) 
        return goal

    def is_success(self, achieved_goal: np.ndarray, desired_goal: np.ndarray) -> Union[np.ndarray, float]:
        d = distance(achieved_goal, desired_goal)
        # print(f"achieved_goal:{achieved_goal}, desired_goal:{desired_goal}, distance:{d}, is finished:{d < self.distance_threshold}")
        return np.array(d < self.distance_threshold, dtype=np.float64)

    def compute_reward(self, achieved_goal, desired_goal, info: Dict[str, Any]) -> Union[np.ndarray, float]:
        d = distance(achieved_goal, desired_goal)
        if self.reward_type == "sparse":
            return -np.array(d > self.distance_threshold, dtype=np.float64)
        else:
            return -d



from typing import Optional

import numpy as np
from gymnasium import spaces

from panda_gym.envs.core import PyBulletRobot
from panda_gym.pybullet import PyBullet


class MyPanda(PyBulletRobot):
    """Panda robot in PyBullet.

    Args:
        sim (PyBullet): Simulation instance.
        block_gripper (bool, optional): Whether the gripper is blocked. Defaults to False.
        base_position (np.ndarray, optionnal): Position of the base base of the robot, as (x, y, z). Defaults to (0, 0, 0).
        control_type (str, optional): "ee" to control end-effector displacement or "joints" to control joint angles.
            Defaults to "ee".
    """

    def __init__(
        self,
        sim: PyBullet,
        block_gripper: bool = False,
        base_position: Optional[np.ndarray] = None,
        control_type: str = "ee",
        action_space: spaces.Box = None
    ) -> None:
        base_position = base_position if base_position is not None else np.zeros(3)
        self.block_gripper = block_gripper
        self.control_type = control_type
        n_action = 3 if self.control_type == "ee" else 7  # control (x, y z) if "ee", else, control the 7 joints
        n_action += 0 if self.block_gripper else 1
        action_space = action_space or spaces.Box(-0.5, 0.5, shape=(n_action,), dtype=np.float32)
        #print(f"action space:{action_space}")
        super().__init__(
            sim,
            body_name="panda",
            file_name="franka_panda/panda.urdf",
            base_position=base_position,
            action_space=action_space,
            joint_indices=np.array([0, 1, 2, 3, 4, 5, 6, 9, 10]),
            joint_forces=np.array([87.0, 87.0, 87.0, 87.0, 12.0, 120.0, 120.0, 170.0, 170.0]),
        )

        self.fingers_indices = np.array([9, 10])
        self.neutral_joint_values = np.array([0.00, 0.41, 0.00, -1.85, 0.00, 2.26, 0.79, 0.00, 0.00])
        # self.neutral_joint_values = np.array([0.00, 0.00, 0.00, -1.85, 0.00, 0.00, 0.00, 0.00, 0.00])
        # self.neutral_joint_values = np.array([1.00, 0.00, 0.00, -1.85, 0.00, 1.00, 0.00, 0.00, 0.4])
        self.ee_link = 11
        self.sim.set_lateral_friction(self.body_name, self.fingers_indices[0], lateral_friction=1.0)
        self.sim.set_lateral_friction(self.body_name, self.fingers_indices[1], lateral_friction=1.0)
        self.sim.set_spinning_friction(self.body_name, self.fingers_indices[0], spinning_friction=0.001)
        self.sim.set_spinning_friction(self.body_name, self.fingers_indices[1], spinning_friction=0.001)

    def set_action(self, action: np.ndarray) -> None:
        action = action.copy()  # ensure action don't change
        action = np.clip(action, self.action_space.low, self.action_space.high)
        if self.control_type == "ee":
            ee_displacement = action[:3]
            target_arm_angles = self.ee_displacement_to_target_arm_angles(ee_displacement)
        else:
            arm_joint_ctrl = action[:7]
            target_arm_angles = self.arm_joint_ctrl_to_target_arm_angles(arm_joint_ctrl)

        if self.block_gripper:
            target_fingers_width = 0
        else:
            fingers_ctrl = action[-1] * 0.2  # limit maximum change in position
            fingers_width = self.get_fingers_width()
            target_fingers_width = fingers_width + fingers_ctrl

        target_angles = np.concatenate((target_arm_angles, [target_fingers_width / 2, target_fingers_width / 2]))
        self.control_joints(target_angles=target_angles)

    def ee_displacement_to_target_arm_angles(self, ee_displacement: np.ndarray) -> np.ndarray:
        """Compute the target arm angles from the end-effector displacement.

        Args:
            ee_displacement (np.ndarray): End-effector displacement, as (dx, dy, dy).

        Returns:
            np.ndarray: Target arm angles, as the angles of the 7 arm joints.
        """
        ee_displacement = ee_displacement[:3] * 0.05  # limit maximum change in position
        # get the current position and the target position
        ee_position = self.get_ee_position()
        target_ee_position = ee_position + ee_displacement
        # Clip the height target. For some reason, it has a great impact on learning
        target_ee_position[2] = np.max((0, target_ee_position[2]))
        # compute the new joint angles
        target_arm_angles = self.inverse_kinematics(
            link=self.ee_link, position=target_ee_position, orientation=np.array([1.0, 0.0, 0.0, 0.0])
        )
        target_arm_angles = target_arm_angles[:7]  # remove fingers angles
        return target_arm_angles

    def arm_joint_ctrl_to_target_arm_angles(self, arm_joint_ctrl: np.ndarray) -> np.ndarray:
        """Compute the target arm angles from the arm joint control.

        Args:
            arm_joint_ctrl (np.ndarray): Control of the 7 joints.

        Returns:
            np.ndarray: Target arm angles, as the angles of the 7 arm joints.
        """
        arm_joint_ctrl = arm_joint_ctrl * 0.05  # limit maximum change in position
        # get the current position and the target position
        current_arm_joint_angles = np.array([self.get_joint_angle(joint=i) for i in range(7)])
        target_arm_angles = current_arm_joint_angles + arm_joint_ctrl
        return target_arm_angles

    def get_obs(self) -> np.ndarray:
        # end-effector position and velocity
        ee_position = np.array(self.get_ee_position())
        ee_velocity = np.array(self.get_ee_velocity())
        # fingers opening
        if not self.block_gripper:
            fingers_width = self.get_fingers_width()
            obs = np.concatenate((ee_position, ee_velocity, [fingers_width]))
        else:
            obs = np.concatenate((ee_position, ee_velocity))
        return obs
    
    # def reset(self,
    #           *,
    #           seed: Optional[int] = None,
    #           options: Optional[dict] = None,
    # ):
    #     # abstract calls our _reset and we add the goal landmark afterwards. only way I found to fit in the framework and create which goal you want.
    #     # Assigning a specific goal_idx will make even a goal-conditioned agent get it in the specific episode instead of randomly picking a goal.
    #     # We want this goal_idx assigning to arrive from 'options' and be assigned for every step of the episode, but 'reset' is called from reset directly and also from 'step'.
    #     # So if 'options' arrived, we assign, as this is the first 'reset'. In case of being called from 'step', options is not given,
    #     # but we don't want to change self.goal_idx since it was correctly put in the earlier 'reset'. So we recognize it in the elif by asking if goal_idx exists,
    #     # and if it does - we leave it as-is: -1 will keep it -1 and use the same goal, and a specified number will keep it too.
    #     # putting None will pick it randomly.
    #     # You expect first time else, second time if, third time (from step) elif.
    #     self.set_joint_neutral()
    #     if options:
    #         self.goal_idx = options["goal_idx"]
    #     elif hasattr(self, 'goal_idx'):
    #         pass
    #     else:
    #         self.goal_idx = None
    #     return super().reset(seed=seed, options=options)

    def reset(self) -> None:
        self.set_joint_neutral()

    def set_joint_neutral(self) -> None:
        """Set the robot to its neutral pose."""
        self.set_joint_angles(self.neutral_joint_values)

    def get_fingers_width(self) -> float:
        """Get the distance between the fingers."""
        finger1 = self.sim.get_joint_angle(self.body_name, self.fingers_indices[0])
        finger2 = self.sim.get_joint_angle(self.body_name, self.fingers_indices[1])
        return finger1 + finger2

    def get_ee_position(self) -> np.ndarray:
        """Returns the position of the ned-effector as (x, y, z)"""
        return self.get_link_position(self.ee_link)

    def get_ee_velocity(self) -> np.ndarray:
        """Returns the velocity of the end-effector as (vx, vy, vz)"""
        return self.get_link_velocity(self.ee_link)



from typing import Optional

from gymnasium import spaces
import numpy as np

from panda_gym.envs.core import RobotTaskEnv
from panda_gym.envs.robots.panda import Panda

from gr_envs.panda_scripts.envs.robots.my_panda import MyPanda

from panda_gym.envs.tasks.flip import Flip
from panda_gym.envs.tasks.pick_and_place import PickAndPlace
from panda_gym.envs.tasks.push import Push
from panda_gym.envs.tasks.reach import Reach

from gr_envs.panda_scripts.envs.tasks.my_reach import MyReach

from panda_gym.envs.tasks.slide import Slide
from panda_gym.envs.tasks.stack import Stack
from panda_gym.pybullet import PyBullet


class PandaFlipEnv(RobotTaskEnv):
    """Pick and Place task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.

    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = Panda(sim, block_gripper=False, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type)
        task = Flip(sim, reward_type=reward_type)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )


class PandaPickAndPlaceEnv(RobotTaskEnv):
    """Pick and Place task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.
    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = Panda(sim, block_gripper=False, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type)
        task = PickAndPlace(sim, reward_type=reward_type)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )


class PandaPushEnv(RobotTaskEnv):
    """Push task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.
    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = Panda(sim, block_gripper=True, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type)
        task = Push(sim, reward_type=reward_type)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )


class PandaReachEnv(RobotTaskEnv):
    """Reach task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.
    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = Panda(sim, block_gripper=True, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type)
        task = Reach(sim, reward_type=reward_type, get_ee_position=robot.get_ee_position)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )

class PandaMyReachEnv(RobotTaskEnv):
    """Reach task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.
    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
        goal_x: int = None, 
        goal_y: int = None, 
        goal_z: int = None, 
        action_space: spaces.Box = None,
        goal_range_negative: float = -0.5,
        goal_range_positive: float = 0.2
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = MyPanda(sim, block_gripper=True, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type, action_space=action_space)
        task = MyReach(sim, reward_type=reward_type, get_ee_position=robot.get_ee_position, goal_x=goal_x, goal_y=goal_y, goal_z=goal_z, goal_range_negative=goal_range_negative, goal_range_positive=goal_range_positive)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )

class PandaSlideEnv(RobotTaskEnv):
    """Slide task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.
    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = Panda(sim, block_gripper=True, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type)
        task = Slide(sim, reward_type=reward_type)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )


class PandaStackEnv(RobotTaskEnv):
    """Stack task wih Panda robot.

    Args:
        render_mode (str, optional): Render mode. Defaults to "rgb_array".
        reward_type (str, optional): "sparse" or "dense". Defaults to "sparse".
        control_type (str, optional): "ee" to control end-effector position or "joints" to control joint values.
            Defaults to "ee".
        renderer (str, optional): Renderer, either "Tiny" or OpenGL". Defaults to "Tiny" if render mode is "human"
            and "OpenGL" if render mode is "rgb_array". Only "OpenGL" is available for human render mode.
        render_width (int, optional): Image width. Defaults to 720.
        render_height (int, optional): Image height. Defaults to 480.
        render_target_position (np.ndarray, optional): Camera targetting this postion, as (x, y, z).
            Defaults to [0., 0., 0.].
        render_distance (float, optional): Distance of the camera. Defaults to 1.4.
        render_yaw (float, optional): Yaw of the camera. Defaults to 45.
        render_pitch (float, optional): Pitch of the camera. Defaults to -30.
        render_roll (int, optional): Rool of the camera. Defaults to 0.
    """

    def __init__(
        self,
        render_mode: str = "rgb_array",
        reward_type: str = "sparse",
        control_type: str = "ee",
        renderer: str = "Tiny",
        render_width: int = 720,
        render_height: int = 480,
        render_target_position: Optional[np.ndarray] = None,
        render_distance: float = 1.4,
        render_yaw: float = 45,
        render_pitch: float = -30,
        render_roll: float = 0,
    ) -> None:
        sim = PyBullet(render_mode=render_mode, renderer=renderer)
        robot = Panda(sim, block_gripper=False, base_position=np.array([-0.6, 0.0, 0.0]), control_type=control_type)
        task = Stack(sim, reward_type=reward_type)
        super().__init__(
            robot,
            task,
            render_width=render_width,
            render_height=render_height,
            render_target_position=render_target_position,
            render_distance=render_distance,
            render_yaw=render_yaw,
            render_pitch=render_pitch,
            render_roll=render_roll,
        )



import gymnasium as gym
from gymnasium.spaces import Box, Dict
import numpy as np

class CombineAchievedGoalAndObservationWrapper(gym.ObservationWrapper):
    def __init__(self, env):
        super(CombineAchievedGoalAndObservationWrapper, self).__init__(env)
        
        # Combine 'achieved_goal' dict and 'observation' box
        achieved_goal_space = env.observation_space['achieved_goal']
        desired_goal_space = env.observation_space['desired_goal']
        observation_space = env.observation_space['observation']
        
        # Add 'observation' space to 'achieved_goal' dict
        combined_spaces = {**achieved_goal_space.spaces, **desired_goal_space.spaces, 'observation': observation_space}
        
        # Set the new combined observation space
        self.observation_space = Dict(combined_spaces)
        self.inner_env = env
        self.is_success_once = False

    # this is called from the observationWrapper parent class only if it is used, and right after 'reset', to change the observation. no need to call the parent class func.
    def observation(self, observation):
        # Add the 'observation' to the 'achieved_goal' dict in the observation
        combined_observation = {**observation['achieved_goal'], 'observation': observation['observation']}
        return combined_observation
    
    # need to override the wrapper class's implementation of 'step' exactly the same, and add the required functionality.
    def step(self, action):
        observation, reward, terminated, truncated, info = self.inner_env.step(action)
        if "success" in info.keys(): self.is_success_once |= info["success"]
        elif "is_success" in info.keys(): self.is_success_once |= info["is_success"]
        elif "step_task_completions" in info.keys(): self.is_success_once |= (len(info["step_task_completions"]) == 1)
        return self.observation(observation), reward, terminated, truncated, info
        


import itertools as itt

import numpy as np

from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Goal, Lava, Floor
from minigrid.minigrid_env import MiniGridEnv


class CustomColorEnv(MiniGridEnv):

    def __init__(self, plan : list = [], size=13, num_crossings=4, obstacle_type=Lava, start_pos = (3, 1), goal_pos =(1, 1), **kwargs):
        self.size = size
        self.num_crossings = num_crossings
        self.obstacle_type = obstacle_type

        if obstacle_type == Lava:
            mission_space = MissionSpace(
                mission_func=lambda: "avoid the lava and get to the green goal square"
            )
        else:
            mission_space = MissionSpace(
                mission_func=lambda: "find the opening and get to the green goal square"
            )

        super().__init__(
            mission_space=mission_space,
            grid_size=size,
            max_steps=4 * size * size,
            # Set this to True for maximum speed
            see_through_walls=False,
            **kwargs
        )
        self.goal_x, self.goal_y = goal_pos
        self.start_pos = start_pos
        self.plan = plan

    def _gen_grid(self, width, height):
        assert width % 2 == 1 and height % 2 == 1  # odd size

        # Create an empty grid
        self.grid = Grid(width, height)

        # Generate the surrounding walls
        self.grid.wall_rect(0, 0, width, height)

        # Place the agent in the top-left corner
        self.agent_pos = np.array(self.start_pos)
        self.agent_dir = 0

        # Place a goal square in the specified location
        self.put_obj(Goal(), self.goal_x, self.goal_y)
        if self.size == 13:
            # room
            self.put_obj(self.obstacle_type(), 7, 1)
            self.put_obj(self.obstacle_type(), 7, 2)
            self.put_obj(self.obstacle_type(), 7, 3)
            self.put_obj(self.obstacle_type(), 7, 4)
            
            self.put_obj(self.obstacle_type(), 11, 4)
            self.put_obj(self.obstacle_type(), 10, 4)
            self.put_obj(self.obstacle_type(), 9, 4)
            # end_of_room
            
            self.put_obj(self.obstacle_type(), 8, 11)
            self.put_obj(self.obstacle_type(), 8, 10)
            self.put_obj(self.obstacle_type(), 8, 9)
            self.put_obj(self.obstacle_type(), 8, 8)

            self.put_obj(self.obstacle_type(), 2, 8)
            self.put_obj(self.obstacle_type(), 3, 8)
            self.put_obj(self.obstacle_type(), 4, 8)
            self.put_obj(self.obstacle_type(), 5, 8)
        
        else:
            self.put_obj(self.obstacle_type(), 2, 1)
            self.put_obj(self.obstacle_type(), 2, 2)
            self.put_obj(self.obstacle_type(), 2, 3)
            self.put_obj(self.obstacle_type(), 2, 4)
            
            self.put_obj(self.obstacle_type(), 5, 7)
            self.put_obj(self.obstacle_type(), 5, 6)
            self.put_obj(self.obstacle_type(), 5, 5)
            self.put_obj(self.obstacle_type(), 5, 4)
            self.put_obj(self.obstacle_type(), 5, 3)
                
            self.mission = (
                "avoid the lava and get to the green goal square"
                if self.obstacle_type == Lava
                else "find the opening and get to the green goal square"
            )
            
        # draw the plan
        for x,y in self.plan:
            if x == self.goal_x and y == self.goal_y:
                continue
            elif (x,y) == self.start_pos:
                self.grid.set(x, y, Floor(color='yellow'))
            else:
                self.grid.set(x, y, Floor(color='purple'))
                



from abc import abstractmethod
from typing import Optional, List

from gymnasium import Env
import numpy as np

from highway_env.envs.common.abstract import AbstractEnv
from highway_env.envs.common.observation import MultiAgentObservation, observation_factory
from highway_env.road.lane import StraightLane, LineType
from highway_env.road.road import Road, RoadNetwork
from highway_env.vehicle.graphics import VehicleGraphics
from highway_env.vehicle.kinematics import Vehicle
from highway_env.vehicle.objects import Landmark, Obstacle


class GoalEnv(Env):
    """
    Interface for A goal-based environment.

    This interface is needed by agents such as Stable Baseline3's Hindsight Experience Replay (HER) agent.
    It was originally part of https://github.com/openai/gym, but was later moved
    to https://github.com/Farama-Foundation/gym-robotics. We cannot add gym-robotics to this project's dependencies,
    since it does not have an official PyPi package, PyPi does not allow direct dependencies to git repositories.
    So instead, we just reproduce the interface here.

    A goal-based environment. It functions just as any regular OpenAI Gym environment but it
    imposes a required structure on the observation_space. More concretely, the observation
    space is required to contain at least three elements, namely `observation`, `desired_goal`, and
    `achieved_goal`. Here, `desired_goal` specifies the goal that the agent should attempt to achieve.
    `achieved_goal` is the goal that it currently achieved instead. `observation` contains the
    actual observations of the environment as per usual.
    """

    @abstractmethod
    def compute_reward(self, achieved_goal: np.ndarray, desired_goal: np.ndarray, info: dict) -> float:
        """Compute the step reward. This externalizes the reward function and makes
        it dependent on a desired goal and the one that was achieved. If you wish to include
        additional rewards that are independent of the goal, you can include the necessary values
        to derive it in 'info' and compute it accordingly.
        Args:
            achieved_goal (object): the goal that was achieved during execution
            desired_goal (object): the desired goal that we asked the agent to attempt to achieve
            info (dict): an info dictionary with additional information
        Returns:
            float: The reward that corresponds to the provided achieved goal w.r.t. to the desired
            goal. Note that the following should always hold true:
                ob, reward, done, info = env.step()
                assert reward == env.compute_reward(ob['achieved_goal'], ob['desired_goal'], info)
        """
        raise NotImplementedError


class ParkingEnv(AbstractEnv, GoalEnv):
    """
    A continuous control environment.

    It implements a reach-type task, where the agent observes their position and speed and must
    control their acceleration and steering so as to reach a given goal.

    Credits to Munir Jojo-Verge for the idea and initial implementation.
    """

    # For parking env with GrayscaleObservation, the env need
    # this PARKING_OBS to calculate the reward and the info.
    # Bug fixed by Mcfly(https://github.com/McflyWZX)
    PARKING_OBS = {"observation": {
            "type": "KinematicsGoal",
            "features": ['x', 'y', 'vx', 'vy', 'cos_h', 'sin_h'],
            "scales": [100, 100, 5, 5, 1, 1],
            "normalize": False
        }}
    DEFAULT_N_SPOTS = 14

    def __init__(self, config: dict = None, render_mode: Optional[str] = None, goal_index: int = -1, n_spots: int = DEFAULT_N_SPOTS, parked_cars: List[int] = []) -> None:
        self.observation_type_parking = None
        self.goal_index = goal_index
        self.n_spots = n_spots
        self.parked_cars = parked_cars
        super().__init__(config, render_mode)

    @classmethod
    def default_config(cls) -> dict:
        config = super().default_config()
        config.update({
            "observation": {
                "type": "KinematicsGoal",
                "features": ['x', 'y', 'vx', 'vy', 'cos_h', 'sin_h'],
                "scales": [100, 100, 5, 5, 1, 1],
                "normalize": False
            },
            "action": {
                "type": "ContinuousAction"
            },
            "reward_weights": [1, 0.3, 0, 0, 0.02, 0.02],
            "success_goal_reward": 0.12,
            "collision_reward": -5,
            "steering_range": np.deg2rad(45),
            "simulation_frequency": 15,
            "policy_frequency": 5,
            "duration": 100,
            "screen_width": 600,
            "screen_height": 300,
            "centering_position": [0.5, 0.5],
            "scaling": 7,
            "controlled_vehicles": 1,
            "vehicles_count": 0,
            "add_walls": True
        })
        return config

    def define_spaces(self) -> None:
        """
        Set the types and spaces of observation and action from config.
        """
        super().define_spaces()
        self.observation_type_parking = observation_factory(self, self.PARKING_OBS["observation"])

    def _info(self, obs, action) -> dict:
        info = super(ParkingEnv, self)._info(obs, action)
        if isinstance(self.observation_type, MultiAgentObservation):
            success = tuple(self._is_success(agent_obs['achieved_goal'], agent_obs['desired_goal']) for agent_obs in obs)
        else:
            obs = self.observation_type_parking.observe()
            success = self._is_success(obs['achieved_goal'], obs['desired_goal'])
        info.update({"is_success": success})
        return info
        
    # reset is called from abstract without goal_idx. this is specifically for env creation for GR generate_observation
    def _reset(self):
        self._create_road(spots=self.n_spots)
        self._create_vehicles()
        # if not hasattr(self, 'goal') or (hasattr(self, 'goal_idx') and self.goal_idx != None): # this is sometimes called from 'step', but the goal already exists. no need to create it again...
        #     self._create_goal_landmark(self.goal_idx)
        # elif hasattr(self, 'goal') and self.goal not in self.road.objects:
        #     self.road.objects.append(self.goal)
        self._create_goal_landmark(self.goal_idx)
        
    def reset(self,
              *,
              seed: Optional[int] = None,
              options: Optional[dict] = None,
    ):
        # abstract calls our _reset and we add the goal landmark afterwards. only way I found to fit in the framework and create which goal you want.
        # Assigning a specific goal_idx will make even a goal-conditioned agent get it in the specific episode instead of randomly picking a goal.
        # We want this goal_idx assigning to arrive from 'options' and be assigned for every step of the episode, but 'reset' is called from reset directly and also from 'step'.
        # So if 'options' arrived, we assign, as this is the first 'reset'. In case of being called from 'step', options is not given,
        # but we don't want to change self.goal_idx since it was correctly put in the earlier 'reset'. So we recognize it in the elif by asking if goal_idx exists,
        # and if it does - we leave it as-is: -1 will keep it -1 and use the same goal, and a specified number will keep it too.
        # putting None will pick it randomly.
        # You expect first time else, second time if, third time (from step) elif.
        if options:
            self.goal_idx = options["goal_idx"]
        elif hasattr(self, 'goal_idx'):
            pass
        else:
            self.goal_idx = None
        return super().reset(seed=seed, options=options)

    def _create_road(self, spots) -> None:
        """
        Create a road composed of straight adjacent lanes.

        :param spots: number of spots in the parking
        """
        net = RoadNetwork()
        width = 4.0
        lt = (LineType.CONTINUOUS, LineType.CONTINUOUS)
        x_offset = 0
        y_offset = 10
        length = 8
        for k in range(spots):
            x = (k + 1 - spots // 2) * (width + x_offset) - width / 2
            net.add_lane("a", "b", StraightLane([x, y_offset], [x, y_offset+length], width=width, line_types=lt))
            net.add_lane("b", "c", StraightLane([x, -y_offset], [x, -y_offset-length], width=width, line_types=lt))

        self.road = Road(network=net,
                         np_random=self.np_random,
                         record_history=self.config["show_trajectories"])
        
    def _create_goal_landmark(self, goal_idx):
        lanes = self.road.network.lanes_list()
        # Goal
        if goal_idx != None: # SHOULD ONLY BE GIVEN FOR A parking-v0 AGENT, NAMELY A GOAL-CONDITIONED AGENT!
            assert self.goal_index == -1 # make sure if a specific goal_idx was given then this is a goal-conditioned agent's env, goal renders randomly every episode.
        elif len(self.road.network.lanes_list()) > self.goal_index >= 0: # in case of a goal-directed agent that has a specific goal_index, it is assigned here and added as a landmark.
            goal_idx = self.goal_index
        else:
            goal_idx = self.np_random.choice(len(lanes))

        lane = lanes[goal_idx]

        self.goal = Landmark(self.road, lane.position(lane.length/2, 0), heading=lane.heading)
        for vehicle in self.controlled_vehicles:
            if hasattr(vehicle, 'goal'):
                pass
            else:
                vehicle.goal = self.goal
        self.road.objects.append(self.goal) # depleted after every step, need to add the goal again at every step.

    def _create_vehicles(self) -> None:
        """Create some new random vehicles of a given type, and add them on the road."""
        # Controlled vehicles
        self.controlled_vehicles = []
        for i in range(self.config["controlled_vehicles"]):
            vehicle = self.action_type.vehicle_class(self.road, [i*20, 0], 2*np.pi*self.np_random.uniform(), 0)
            vehicle.color = VehicleGraphics.EGO_COLOR
            self.road.vehicles.append(vehicle)
            self.controlled_vehicles.append(vehicle)

        for parked_vehicle_spot in self.parked_cars:
            lane = ("a", "b", parked_vehicle_spot) if parked_vehicle_spot <= self.n_spots else ("b", "c", self.n_spots - parked_vehicle_spot)
            v = Vehicle.make_on_lane(self.road, lane, 4, speed=0)
            self.road.vehicles.append(v)

        # Other vehicles
        for i in range(self.config["vehicles_count"]):
            lane = ("a", "b", i) if self.np_random.uniform() >= 0.5 else ("b", "c", i)
            v = Vehicle.make_on_lane(self.road, lane, 4, speed=0)
            self.road.vehicles.append(v)
        # for v in self.road.vehicles:  # Prevent early collisions
        #     if v is not self.vehicle and (
        #             np.linalg.norm(v.position - self.goal.position) < 20 or
        #             np.linalg.norm(v.position - self.vehicle.position) < 20):
        #         self.road.vehicles.remove(v)

        # Walls
        if self.config["add_walls"]:
            width, height = 70, 42
            for y in [-height / 2, height / 2]:
                obstacle = Obstacle(self.road, [0, y])
                obstacle.LENGTH, obstacle.WIDTH = (width, 1)
                obstacle.diagonal = np.sqrt(obstacle.LENGTH**2 + obstacle.WIDTH**2)
                self.road.objects.append(obstacle)
            for x in [-width / 2, width / 2]:
                obstacle = Obstacle(self.road, [x, 0], heading=np.pi / 2)
                obstacle.LENGTH, obstacle.WIDTH = (height, 1)
                obstacle.diagonal = np.sqrt(obstacle.LENGTH**2 + obstacle.WIDTH**2)
                self.road.objects.append(obstacle)

    def compute_reward(self, achieved_goal: np.ndarray, desired_goal: np.ndarray, info: dict, p: float = 0.5) -> float:
        """
        Proximity to the goal is rewarded

        We use a weighted p-norm

        :param achieved_goal: the goal that was achieved
        :param desired_goal: the goal that was desired
        :param dict info: any supplementary information
        :param p: the Lp^p norm used in the reward. Use p<1 to have high kurtosis for rewards in [0, 1]
        :return: the corresponding reward
        """
        return -np.power(np.dot(np.abs(achieved_goal - desired_goal), np.array(self.config["reward_weights"])), p)

    def _reward(self, action: np.ndarray) -> float:
        obs = self.observation_type_parking.observe()
        obs = obs if isinstance(obs, tuple) else (obs,)
        reward = sum(self.compute_reward(agent_obs['achieved_goal'], agent_obs['desired_goal'], {}) for agent_obs in obs)
        reward += self.config['collision_reward'] * sum(v.crashed for v in self.controlled_vehicles)
        return reward

    def _is_success(self, achieved_goal: np.ndarray, desired_goal: np.ndarray) -> bool:
        return self.compute_reward(achieved_goal, desired_goal, {}) > -self.config["success_goal_reward"]

    def _is_terminated(self) -> bool:
        """The episode is over if the ego vehicle crashed or the goal is reached or time is over."""
        crashed = any(vehicle.crashed for vehicle in self.controlled_vehicles)
        obs = self.observation_type_parking.observe()
        obs = obs if isinstance(obs, tuple) else (obs,)
        success = all(self._is_success(agent_obs['achieved_goal'], agent_obs['desired_goal']) for agent_obs in obs)
        return bool(crashed or success)

    def _is_truncated(self) -> bool:
        """The episode is truncated if the time is over."""
        return self.time >= self.config["duration"]


class ParkingEnvActionRepeat(ParkingEnv):
    def __init__(self):
        super().__init__({"policy_frequency": 1, "duration": 20})


class ParkingEnvParkedVehicles(ParkingEnv):
    def __init__(self):
        super().__init__({"vehicles_count": 10})

        
# class MultiTaskParkingEnv(AbstractEnv, GoalEnv):
#     """
#     A continuous control environment.

#     It implements a reach-type task, where the agent observes their position and speed and must
#     control their acceleration and steering so as to reach a given goal.

#     Credits to Munir Jojo-Verge for the idea and initial implementation.
#     """

#     # For parking env with GrayscaleObservation, the env need
#     # this PARKING_OBS to calculate the reward and the info.
#     # Bug fixed by Mcfly(https://github.com/McflyWZX)
#     PARKING_OBS = {"observation": {
#             "type": "KinematicsGoal",
#             "features": ['x', 'y', 'vx', 'vy', 'cos_h', 'sin_h'],
#             "scales": [100, 100, 5, 5, 1, 1],
#             "normalize": False
#         }}
#     DEFAULT_N_SPOTS = 14

#     def __init__(self, config: dict = None, render_mode: Optional[str] = None, goal_indices: List = [-1], n_spots: int = DEFAULT_N_SPOTS, parked_cars: List[int] = []) -> None:
#         self.observation_type_parking = None
#         self.goal_indices = goal_indices
#         self.n_spots = n_spots
#         self.parked_cars = parked_cars
#         super().__init__(config, render_mode)

#     @classmethod
#     def default_config(cls) -> dict:
#         config = super().default_config()
#         config.update({
#             "observation": {
#                 "type": "KinematicsGoal",
#                 "features": ['x', 'y', 'vx', 'vy', 'cos_h', 'sin_h'],
#                 "scales": [100, 100, 5, 5, 1, 1],
#                 "normalize": False
#             },
#             "action": {
#                 "type": "ContinuousAction"
#             },
#             "reward_weights": [1, 0.3, 0, 0, 0.02, 0.02],
#             "success_goal_reward": 0.12,
#             "collision_reward": -5,
#             "steering_range": np.deg2rad(45),
#             "simulation_frequency": 15,
#             "policy_frequency": 5,
#             "duration": 100,
#             "screen_width": 600,
#             "screen_height": 300,
#             "centering_position": [0.5, 0.5],
#             "scaling": 7,
#             "controlled_vehicles": 1,
#             "vehicles_count": 0,
#             "add_walls": True
#         })
#         return config

#     def define_spaces(self) -> None:
#         """
#         Set the types and spaces of observation and action from config.
#         """
#         super().define_spaces()
#         self.observation_type_parking = observation_factory(self, self.PARKING_OBS["observation"])

#     def _info(self, obs, action) -> dict:
#         info = super(ParkingEnv, self)._info(obs, action)
#         if isinstance(self.observation_type, MultiAgentObservation):
#             success = tuple(self._is_success(agent_obs['achieved_goal'], agent_obs['desired_goal']) for agent_obs in obs)
#         else:
#             obs = self.observation_type_parking.observe()
#             success = self._is_success(obs['achieved_goal'], obs['desired_goal'])
#         info.update({"is_success": success})
#         return info

#     def _reset(self):
#         self._create_road(spots=self.n_spots)
#         self._create_vehicles()

#     def _create_road(self, spots) -> None:
#         """
#         Create a road composed of straight adjacent lanes.

#         :param spots: number of spots in the parking
#         """
#         net = RoadNetwork()
#         width = 4.0
#         lt = (LineType.CONTINUOUS, LineType.CONTINUOUS)
#         x_offset = 0
#         y_offset = 10
#         length = 8
#         for k in range(spots):
#             x = (k + 1 - spots // 2) * (width + x_offset) - width / 2
#             net.add_lane("a", "b", StraightLane([x, y_offset], [x, y_offset+length], width=width, line_types=lt))
#             net.add_lane("b", "c", StraightLane([x, -y_offset], [x, -y_offset-length], width=width, line_types=lt))

#         self.road = Road(network=net,
#                          np_random=self.np_random,
#                          record_history=self.config["show_trajectories"])

#     def _create_vehicles(self) -> None:
#         """Create some new random vehicles of a given type, and add them on the road."""
#         # Controlled vehicles
#         self.controlled_vehicles = []
#         for i in range(self.config["controlled_vehicles"]):
#             vehicle = self.action_type.vehicle_class(self.road, [i*20, 0], 2*np.pi*self.np_random.uniform(), 0)
#             vehicle.color = VehicleGraphics.EGO_COLOR
#             self.road.vehicles.append(vehicle)
#             self.controlled_vehicles.append(vehicle)

#         lanes = self.road.network.lanes_list()
#         # Goal
#         if any(len(self.road.network.lanes_list()) < element for element in goal_index):
#             # we don't support a goal-conditioned agent here: only a multi-task agent with specific goal tasks. Thus we won't generate random goals for an episode if a goal_index of -1 was given.
#             raise NotImplementedError(f"Can't use network lanes larger than {len(self.road.network.lanes_list())}")
#         else:
#             goal_indices = self.goal_indices
#             self.achieved_goals_indices = []
            
#         goal_lanes = [lanes[goal_idx] for goal_idx in goal_indices]

#         self.goals = [Landmark(self.road, lane.position(lane.length/2, 0), heading=lane.heading) for lane in goal_lanes]
#         for goal in self.goals:
#             self.road.objects.append(goal)

#         for parked_vehicle_spot in self.parked_cars:
#             lane = ("a", "b", parked_vehicle_spot) if parked_vehicle_spot <= self.n_spots else ("b", "c", self.n_spots - parked_vehicle_spot)
#             v = Vehicle.make_on_lane(self.road, lane, 4, speed=0)
#             self.road.vehicles.append(v)

#         # Other vehicles
#         for i in range(self.config["vehicles_count"]):
#             lane = ("a", "b", i) if self.np_random.uniform() >= 0.5 else ("b", "c", i)
#             v = Vehicle.make_on_lane(self.road, lane, 4, speed=0)
#             self.road.vehicles.append(v)
#         # for v in self.road.vehicles:  # Prevent early collisions
#         #     if v is not self.vehicle and (
#         #             np.linalg.norm(v.position - self.goal.position) < 20 or
#         #             np.linalg.norm(v.position - self.vehicle.position) < 20):
#         #         self.road.vehicles.remove(v)

#         # Walls
#         if self.config['add_walls']:
#             width, height = 70, 42
#             for y in [-height / 2, height / 2]:
#                 obstacle = Obstacle(self.road, [0, y])
#                 obstacle.LENGTH, obstacle.WIDTH = (width, 1)
#                 obstacle.diagonal = np.sqrt(obstacle.LENGTH**2 + obstacle.WIDTH**2)
#                 self.road.objects.append(obstacle)
#             for x in [-width / 2, width / 2]:
#                 obstacle = Obstacle(self.road, [x, 0], heading=np.pi / 2)
#                 obstacle.LENGTH, obstacle.WIDTH = (height, 1)
#                 obstacle.diagonal = np.sqrt(obstacle.LENGTH**2 + obstacle.WIDTH**2)
#                 self.road.objects.append(obstacle)

#     def compute_reward(self, achieved_goal: np.ndarray, desired_goals: np.ndarray, info: dict, p: float = 0.5) -> float:
#         """
#         Achievement of each goal is rewarded

#         :param achieved_goal: the goal that was achieved
#         :param desired_goal: the goal that was desired
#         :param dict info: any supplementary information
#         :param p: the Lp^p norm used in the reward. Use p<1 to have high kurtosis for rewards in [0, 1]
#         :return: the corresponding reward
#         """
#         if any((-np.power(np.dot(np.abs(achieved_goal - desired_goal), np.array(self.config["reward_weights"])), p)) > -self.config["success_goal_reward"] for desired_goal in desired_goals): return 1
#         else: return 0

#     # only relevant for multiagents
#     def _reward(self, action: np.ndarray) -> float:
#         obs = self.observation_type_parking.observe()
#         obs = obs if isinstance(obs, tuple) else (obs,)
#         reward = sum(self.compute_reward(agent_obs['achieved_goal'], agent_obs['desired_goal'], {}) for agent_obs in obs)
#         reward += self.config['collision_reward'] * sum(v.crashed for v in self.controlled_vehicles)
#         return reward

#     def _is_success(self, achieved_goal: np.ndarray, desired_goals: np.ndarray) -> bool:
#         return self.compute_reward(achieved_goal, desired_goals, {}) > 0

#     def _is_terminated(self) -> bool:
#         """The episode is over if the ego vehicle crashed or the goal is reached or time is over."""
#         crashed = any(vehicle.crashed for vehicle in self.controlled_vehicles)
#         obs = self.observation_type_parking.observe()
#         obs = obs if isinstance(obs, tuple) else (obs,)
#         success = all(self._is_success(agent_obs['achieved_goal'], agent_obs['desired_goal']) for agent_obs in obs)
#         return bool(crashed or success)

#     def _is_truncated(self) -> bool:
#         """The episode is truncated if the time is over."""
#         return self.time >= self.config["duration"]
